-- 	DynamicAttach Script
--
-- 	Author: 	Heady
-- 	Convert:	Silas770
-- 	Date:		27.03.2019
--	Version:	1.1.0.0 (FS19)
--	

local modName = g_currentModName;

DynamicAttach = {};

function DynamicAttach.prerequisitesPresent(specializations)
    return true;
end;

function DynamicAttach.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "findDynamicAttachObjectCallback", DynamicAttach.findDynamicAttachObjectCallback);
	SpecializationUtil.registerFunction(vehicleType, "setDynamicAttach", DynamicAttach.setDynamicAttach);
end;

function DynamicAttach.registerEventListeners(vehicleType)
	local els = {"onLoad", "onReadStream", "onWriteStream", "onDraw", "onUpdateTick", "onRegisterActionEvents"};
	for _, el in pairs(els) do
		SpecializationUtil.registerEventListener(vehicleType, el, DynamicAttach);
	end;
end;

function DynamicAttach:onLoad(savegame)
	self.spec_dynamicAttach = self["spec_" .. modName .. ".DynamicAttach"];
	local spec = self.spec_dynamicAttach;
	
	spec.attached = false;
	spec.dirty = false;
	
	spec.objectsInRange = {};
	spec.objectsToJoint = {};
	
	spec.raycastNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.dynamicAttach#raycastNode"), self.i3dMappings);
	spec.boxSizeX = getXMLFloat(self.xmlFile, "vehicle.dynamicAttach#boxSizeX", 1);
	spec.boxSizeY = getXMLFloat(self.xmlFile, "vehicle.dynamicAttach#boxSizeY", 1);
	spec.boxSizeZ = getXMLFloat(self.xmlFile, "vehicle.dynamicAttach#boxSizeZ", 1);
	
	if spec.raycastNode == nil then
		g_logManager:xmlWarning(self.configFileName, "'dynamicAttach#raycastNode' is not defined - DynamicAttach will not work")
	end;
end;

function DynamicAttach:onReadStream(streamId, connection)
	local spec = self.spec_dynamicAttach;
	spec.attached = streamReadBool(streamId);
end;

function DynamicAttach:onWriteStream(streamId, connection)
	local spec = self.spec_dynamicAttach;
	streamWriteBool(streamId, spec.attached);
end;

function DynamicAttach:onDraw(isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	if isActiveForInputIgnoreSelection and isSelected then
		local spec = self.spec_dynamicAttach;
	
		local actionEvent = spec.actionEvents[InputAction.DYNAMICATTACH_ATTACH];
		if actionEvent ~= nil then
			local text;
			
			if spec.attached then
				text = "DYNAMICATTACH_DETACH";
			else
				text = "DYNAMICATTACH_ATTACH";
			end;
			
			g_inputBinding:setActionEventText(actionEvent.actionEventId, g_i18n:getText(text));
		end;
	end;
end;

function DynamicAttach:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	local spec = self.spec_dynamicAttach;

	if spec.dirty and spec.raycastNode ~= nil then
		if self.isServer then
			spec.objectsInRange = {};
			local x,y,z = getWorldTranslation(spec.raycastNode);
			local rx,ry,rz = getWorldRotation(spec.raycastNode);
		
			local sizeX, sizeY, sizeZ = spec.boxSizeX, spec.boxSizeY, spec.boxSizeZ;
			overlapBox(x, y, z, rx, ry, rz, sizeX, sizeY, sizeZ, "findDynamicAttachObjectCallback", self, nil, true, false, true);
		
			if spec.attached then
				-- create joints
				for index, object in pairs(spec.objectsInRange) do
					if spec.objectsToJoint[index] == nil then
						local constr = JointConstructor:new();
						constr:setActors(self.components[1].node, object.physics);
						-- create joint at center of mass to avoid jittering (e.g. tree trunks pivots are not at center of mass position)
						local x,y,z = localToWorld(object.physics, getCenterOfMass(object.physics));
						constr:setJointWorldPositions(x,y,z, x,y,z);
						constr:setRotationLimit(0, 0, 0);
						constr:setRotationLimit(1, 0, 0);
						constr:setRotationLimit(2, 0, 0);
						object.dynamicAttacherJoint = constr:finalize();
						spec.objectsToJoint[index] = object;
					end;
				end;
			else
				-- remove joints
				for index, object in pairs(spec.objectsToJoint) do
					if spec.objectsToJoint[index] ~= nil then
						removeJoint(object.dynamicAttacherJoint);
						spec.objectsToJoint[index] = nil;
					end;
				end;
			end;
		end;
		
		spec.dirty = false;
	end;
end;

function DynamicAttach:findDynamicAttachObjectCallback(transformId)
	if transformId ~= 0 and getHasClassId(transformId, ClassIds.SHAPE) then
		local spec = self.spec_dynamicAttach;
		
        local object = g_currentMission:getNodeObject(transformId);
        if object ~= nil then
            if object.getMeshNodes ~= nil and object.dynamicMountObject == nil and spec.objectsInRange[transformId] == nil then
                local nodes = object:getMeshNodes();
                if nodes ~= nil then
                    spec.objectsInRange[transformId] = {physics=object.nodeId, visuals=nodes, object=object};
                end;
            end;
        elseif getSplitType(transformId) ~= 0 then
            local rigidBodyType = getRigidBodyType(transformId);
            if (rigidBodyType == "Dynamic" or rigidBodyType == "Kinematic") and spec.objectsInRange[transformId] == nil then
                spec.objectsInRange[transformId] = {physics=transformId, visuals={transformId}};
            end;
        end;
    end;
	
    return true;
end;

function DynamicAttach:setDynamicAttach(attached, noEventSend)
	local spec = self.spec_dynamicAttach;

	if attached ~= spec.attached then
		spec.attached = not spec.attached;
		spec.dirty = true;

		DynamicAttachEvent.sendEvent(self, spec.attached, noEventSend);
	end;
end;

function DynamicAttach:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
	if self.isClient then
		local spec = self.spec_dynamicAttach;
		self:clearActionEventsTable(spec.actionEvents);
		
		if isActiveForInput then
			local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.DYNAMICATTACH_ATTACH, self, DynamicAttach.actionEventDynamicAttach, false, true, false, true, nil)
            g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_HIGH);
		end;
	end;
end;

function DynamicAttach.actionEventDynamicAttach(self, actionName, inputValue, callbackState, isAnalog)
	local spec = self.spec_dynamicAttach;
	local newState = not spec.attached;
	
	self:setDynamicAttach(newState);
end;


-- MP Event
DynamicAttachEvent = {};
DynamicAttachEvent_mt = Class(DynamicAttachEvent, Event);

InitEventClass(DynamicAttachEvent, "DynamicAttachEvent");

function DynamicAttachEvent:emptyNew()
    local self = Event:new(DynamicAttachEvent_mt);
    self.className = "DynamicAttachEvent";
    return self;
end;

function DynamicAttachEvent:new(object, attached)
    local self = DynamicAttachEvent:emptyNew()
    self.object = object;
	self.attached = attached;
    return self;
end;

function DynamicAttachEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId);
    self.attached = streamReadBool(streamId);
    self:run(connection);
end;

function DynamicAttachEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object);
    streamWriteBool(streamId, self.attached);
end;

function DynamicAttachEvent:run(connection)
    if not connection:getIsServer() then
        g_server:broadcastEvent(self, nil, connection, self.object);
    end;
	if self.object ~= nil then
		self.object:setDynamicAttach(self.attached, true);
	end;
end;

function DynamicAttachEvent.sendEvent(vehicle, attached, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(DynamicAttachEvent:new(vehicle, attached), nil, nil, vehicle);
		else
			g_client:getServerConnection():sendEvent(DynamicAttachEvent:new(vehicle, attached));
		end;
	end;
end;